! ===========================================================================
MODULE tlike

! This code calculates the likelihood function for TT data at low l 
!
!  It assumes that there is no noise
!
!  It uses a smoothed ILC map as the temperature map and the Kp2 sky mask
!
! O. Dore
! D. Spergel, December 2005
!
! ===========================================================================
INTEGER :: ngood,nsmax
DOUBLE PRECISION, ALLOCATABLE :: wl(:),pseudo_cl(:)
DOUBLE PRECISION, ALLOCATABLE :: t_map(:)
DOUBLE PRECISION, ALLOCATABLE :: f_map(:) !difference between ILC and V1
DOUBLE PRECISION, ALLOCATABLE :: MPlM(:,:,:)
DOUBLE PRECISION, ALLOCATABLE :: C0(:,:)  ! C0 computed for clfiducial(nlmax:3*2**ires) + monopole + dipole
DOUBLE PRECISION :: fore_marg
	! fore_marg = 0. no marginalization
	! fore_marg = 1.  (assume that the uncertainty is the diff between
	!			V1 and ILC)
	! fore_marg = 1.e6  (no prior on amplitude of foreground unc.)
	!

CONTAINS


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE SETUP_FOR_TT_EXACT(nlhigh)
!
! DEGRADES Kp2 mask to ires
! SMOOTHES ILC map with a beam exp(-l*(l+1)/(2*lc*(lc+1))!
!	with lc =1.5*2**ires
! COMPUTES C0 for fiducial 
!
use alm_tools
use fitstools
use pix_tools
use wmap_util
use wmap_options, only : WMAP_data_dir
IMPLICIT NONE

include 'read_archive_map.fh'
INTEGER, INTENT(IN) :: nlhigh ! vary cl's from 2 : nlhigh
integer :: ires                 = 3      ! resolution for low l TT code
INTEGER :: npix,np,status,nlmax,m, lun
INTEGER :: ip,jp,ipix,l,ll,lcut,jpix,ifact
CHARACTER*120:: filename
LOGICAL :: exist
INTEGER, ALLOCATABLE :: good(:)
DOUBLE PRECISION, ALLOCATABLE :: w8ring(:,:),window_th(:,:)
REAL, ALLOCATABLE ::cl_fiducial(:),t_ilc(:),dw8(:,:),dummy(:),t_ilc_v1(:),v1(:),map(:)
REAL, ALLOCATABLE :: mask(:),mask_hr(:),t_ring(:),v1_yr(:)
REAL, ALLOCATABLE:: Plxy(:,:,:),cl(:)
COMPLEX, ALLOCATABLE :: alm(:,:,:)
DOUBLE PRECISION :: vec_ip(3),vec_jp(3),x,p(0:512)
!
!	READ IN cl's for fiducial model (used for high l)
!
ALLOCATE(cl_fiducial(2:512))
filename = trim(WMAP_data_dir)//'wmap_lcdm_pl_model_yr1_v1.dat'
INQUIRE(file=filename,exist=exist)
IF(.not.exist) THEN
	write(*,*) ' unable to open cl.dat'
	STOP
ENDIF
call get_free_lun( lun )
OPEN(lun,file=filename,status='unknown')
do l = 2,512
	read(lun,*) ll,cl_fiducial(l)
	cl_fiducial(ll) = cl_fiducial(ll)*1.d-6*2.*3.1415927/(ll*(ll+1))
end do
CLOSE(lun)

!
! READ IN MASK AND DEGRADE
!
nsmax = 2**ires
nlmax = 4*nsmax
npix = 12*nsmax*nsmax
ALLOCATE(mask_hr(0:12*512*512-1))
ALLOCATE(dummy(0:12*512*512-1))
ALLOCATE(mask(0:12*4**ires-1))

filename =trim(WMAP_data_dir)//'wmap_kp2_mask_2yr_v2p3.fits'
call read_archive_map(filename,dummy,mask_hr,np,status)
if(status.ne.0.or.np.ne.12*512*512) then
	write(*,*) ' error reading mask'
	stop
endif
mask = 0
ifact = (512/(2**ires))**2

do ipix = 0,12*4**ires-1
	do jpix = ipix*ifact,(ipix+1)*ifact-1
		mask(ipix) = mask(ipix)+mask_hr(jpix)
	end do
	if(mask(ipix).gt.ifact/2) then
		mask(ipix) = 1
		mask_hr(ipix*ifact:(ipix+1)*ifact-1) = 1
	else
		mask(ipix) = 0
		mask_hr(ipix*ifact:(ipix+1)*ifact-1) = 0
	end if
end do

ngood = sum(mask)

ALLOCATE(good(0:ngood-1))
ip = 0
do ipix = 0,12*nsmax*nsmax-1
	if(mask(ipix).eq.1) then
		good(ip) = ipix
		ip = ip+1
	endif
end do
if(ip.ne.ngood) STOP
DEALLOCATE(dummy)
!
!
!  READ IN ILC MAP AND SMOOTH IT
!	set smoothing scale so that there is minimal power for l > 3*2**ires
!

ALLOCATE(wl(2:1024))
ALLOCATE(window_th(0:nlmax,1))
ALLOCATE(dummy(0:12*512*512-1))
filename='ilc.map'
INQUIRE(file=filename,exist=exist)
ALLOCATE(t_ilc(0:12*512*512-1))
ALLOCATE(v1_yr(0:12*512*512-1))
ALLOCATE(v1(0:12*512*512-1))
ALLOCATE(t_ilc_v1(0:12*512*512-1))
filename = trim(WMAP_data_dir)//'pass2f_fltFiltBZFL_fs_spm_r9_ximb_xsd_yr1_itf.V1.map_i'
call read_archive_map(filename,v1_yr,dummy,np,status) !v1 map
v1 = v1_yr/3.
filename = trim(WMAP_data_dir)//'pass2f_fltFiltBZFL_fs_spm_r9_ximb_xsd_yr2_itf.V1.map_i'
call read_archive_map(filename,v1,dummy,np,status) !v1 map
v1 = v1+v1_yr/3.
filename = trim(WMAP_data_dir)//'pass2f_fltFiltBZFL_fs_spm_r9_ximb_xsd_yr3_itf.V1.map_i'
call read_archive_map(filename,v1,dummy,np,status) !v1 map
v1 = v1+v1_yr/3.
filename = trim(WMAP_data_dir)//'bias_corrected_pass2_3yr_ilc.fits'
call read_archive_map(filename,t_ilc,dummy,np,status)
t_ilc_v1 = t_ilc - v1  !ILC - v1


ALLOCATE(map(0:12*512*512-1))
map = t_ilc
DEALLOCATE(t_ilc)

!---------------------------------------

DEALLOCATE(v1_yr,v1)
ALLOCATE(alm(1,0:512,0:512))
ALLOCATE(w8ring(1:1024,1))
ALLOCATE(dw8(1:1024,1))

filename = trim(WMAP_data_dir)//'healpix_data/weight_ring_n00512.fits'
call input_map(filename,dw8,1024,1)
w8ring = 1.d0+dw8
ALLOCATE(t_ring(0:12*512*512-1))
call convert_map(map,t_ring,512)
call map2alm(512,512,512,t_ring,alm,-1.d0,w8ring)
do l = 2,512
	alm(1,l,0:l) = alm(1,l,0:l)*exp(-1.5*l*(l+1)/(nlmax*nlmax))
end do
call alm2map(512,512,512,alm,t_ring)
call convert_map_ring(t_ring,map,512)
call convert_map(t_ilc_v1,t_ring,512)
call map2alm(512,512,512,t_ring,alm,-1.d0,w8ring)
do l = 2,512
	alm(1,l,0:l) = alm(1,l,0:l)*exp(-1.5*l*(l+1)/(nlmax*nlmax))
end do
call alm2map(512,512,512,alm,t_ring)
call convert_map_ring(t_ring,t_ilc_v1,512)
DEALLOCATE(alm,w8ring,dw8,dummy,t_ring)
map = map*mask_hr 
t_ilc_v1 = t_ilc_v1*mask_hr 
if(status.ne.0.or.np.ne.12*512*512) then
		write(*,*) ' error reading ilc map'
		stop
endif
!
CALL GET_PIXEL_WINDOW(2**ires,window_th)
do l = 2,nlmax
	wl(l) = window_th(l,1)*exp(-1.5*l*(l+1)/(nlmax*nlmax))
end do

!
!  COMPUTE MPLM
!
ALLOCATE(MPlM(2:nlhigh,0:ngood-1,0:ngood-1))
ALLOCATE(C0(0:ngood-1,0:ngood-1))
ALLOCATE(t_map(0:ngood-1))
ALLOCATE(f_map(0:ngood-1))

do ip = 0,ngood-1
	CALL pix2vec_nest(nsmax,good(ip),vec_ip)
	t_map(ip) = sum(map(good(ip)*ifact:(good(ip)+1)*ifact-1))&
		/sum(mask_hr(good(ip)*ifact:(good(ip)+1)*ifact-1))
	f_map(ip) = sum(t_ilc_v1(good(ip)*ifact:(good(ip)+1)*ifact-1))&
		/sum(mask_hr(good(ip)*ifact:(good(ip)+1)*ifact-1))
	do jp = ip,ngood-1
		CALL pix2vec_nest(nsmax,good(jp),vec_jp)
		x = sum(vec_ip(:)*vec_jp(:))
		p(0) = 1.
		p(1) = x
		do l = 1,nlmax-1
			p(l+1) = ((2*l+1)*x*p(l)-l*p(l-1))/(l+1.)
		end do
		do l = 0,nlmax
			p(l) = p(l)*(2*l+1)/(4.*3.1415927)
		end do
		MPlM(2:nlhigh,ip,jp) = p(2:nlhigh)
		C0(ip,jp) = sum(cl_fiducial(nlhigh+1:nlmax)*&
			wl(nlhigh+1:nlmax)**2*p(nlhigh+1:nlmax)) +&
			cl_fiducial(2)*(p(0)+p(1))
		C0(jp,ip) = C0(ip,jp)
	end do
end do

DEALLOCATE(good,map,mask_hr,t_ilc_v1,window_th)
RETURN
END SUBROUTINE SETUP_FOR_TT_EXACT

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


SUBROUTINE COMPUTE_LIKE(nlhigh,cl_in,chisq,lndet) !like)
!
!  This subroutine returns the likelihood 
!
!
!
!
use wmap_util
IMPLICIT NONE
INTEGER, INTENT(IN) :: nlhigh
REAL, INTENT(IN), DIMENSION(2:nlhigh) :: cl_in
!
DOUBLE PRECISION,INTENT(OUT) :: chisq,lndet !,like
!
DOUBLE PRECISION, ALLOCATABLE :: cl(:),c_inv_t_map(:),c(:,:),c_inv_f_map(:)
DOUBLE PRECISION:: fCinvf,like
INTEGER :: info,ip,jp, l
!
ALLOCATE(cl(2:nlhigh))
ALLOCATE(C(0:ngood-1,0:ngood-1))
!
C = C0

do l =2,nlhigh
	cl(l) = cl_in(l)*wl(l)**2*1.e-6*2.*3.1415927/(l*(l+1))
end do
do ip=0,ngood-1
     do jp=0,ngood-1
           do l=2,nlhigh
              C(ip,jp) = C(ip,jp)+cl(l)*MPlM(l,ip,jp)
           enddo
     enddo
enddo
CALL DPOTRF('U',ngood,C,ngood,info) 
IF(info.NE.0)then
!     PRINT*, ' info in dpotrf =', info
!     stop
   call wmap_likelihood_error( 'tlike dpotrf failed', info )
   chisq = 0d0
   lndet = 0d0
   return
endif
lndet = 0.
do ip=0,ngood-1
	lndet = lndet + 2.*log(C(ip,ip)*4.*3.145927/float(12*nsmax*nsmax))
end do
ALLOCATE(c_inv_t_map(0:ngood-1))
ALLOCATE(c_inv_f_map(0:ngood-1))
c_inv_t_map = t_map
call DPOTRS('U', ngood, 1, C, ngood, c_inv_t_map, ngood, INFO)
IF(info.NE.0)then
!     PRINT*, ' info in dpotrs =', info
!     stop
	call wmap_likelihood_error( 'tlike 2nd dpotrs failed', info )
	chisq = 0d0
	lndet = 0d0
	return
endif
c_inv_f_map = f_map
call DPOTRS('U', ngood, 1, C, ngood, c_inv_f_map, ngood, INFO)
IF(info.NE.0)then
!     PRINT*, ' info in dpotrs =', info
!     stop
	call wmap_likelihood_error( 'tlike 1st dpotrs failed', info )
	chisq = 0d0
	lndet = 0d0
	return
endif
fCinvf = sum(c_inv_f_map(:)*f_map(:))/ngood

chisq = sum(c_inv_t_map(:)*t_map(:)) 
if(fore_marg.ne.0) then
	chisq = chisq- (sum(c_inv_f_map(:)*t_map(:)))**2&
			/ngood/(1./fore_marg+fCinvf)
	lndet = lndet + log(1./fore_marg+fCinvf)
endif

like = (chisq+lndet)/2.d0
chisq=chisq/2.
lndet=lndet/2.



DEALLOCATE(cl,C,c_inv_t_map)

return

END subroutine COMPUTE_LIKE



SUBROUTINE GET_PIXEL_WINDOW(nsmax,window_th)
USE fitstools, ONLY : read_dbintab
use wmap_options, only : WMAP_data_dir
use healpix_types
IMPLICIT NONE
INTEGER, INTENT(IN) :: nsmax
DOUBLE PRECISION, INTENT(OUT) :: window_th(0:4*nsmax,1:1)
!
CHARACTER(LEN=80), DIMENSION(1:60)  :: header_PS
CHARACTER(LEN=128)                  :: infile
CHARACTER(LEN=128)                  :: windowfile
CHARACTER(LEN=128)                  :: windowname
CHARACTER(LEN=128), DIMENSION(4)    :: winfiledir
CHARACTER(LEN=4)                  :: sstr
DOUBLE PRECISION :: null_val
LOGICAL(LGT) :: bad,ok
INTEGER :: i
!
write (sstr,'(I4)') nsmax
sstr=adjustl(sstr)
windowname='pixel_window_n'&
       & //repeat('0',4-len_trim(sstr))//trim(sstr)//'.fits'

winfiledir(1)=''

winfiledir(2)=trim(WMAP_data_dir)//'healpix_data/'
winfiledir(3)='./data/'
winfiledir(4)='./'
do i=1,4
   windowfile=trim(winfiledir(i))//trim(windowname)
 
   inquire(file=windowfile, exist=ok)
   if (ok) exit
enddo
if (.not.ok) stop 'Could not find window file.' 
call read_dbintab(windowfile,window_th,4*nsmax+1,1,null_val,bad)

RETURN
END SUBROUTINE GET_PIXEL_WINDOW

SUBROUTINE generate_ylm(nsmax,nlmax,ylm)
!
! This subroutine calculates temperature spherical harmonics
!
!
! Since
!    Y^*_{lm} = (-1)^m Y_{l-m}
! this routine computes them for only m>=0.
!
! y_lm's are normalized such that SUM_i y^*_i y_i = Omega_pix^{-1},
! which means that
!
! \int dx y^*_{lm}(x) y_{l'm'}(x) = delta_{ll'}delta_{mm'}
!
! and
!
! Sum_{l=0}Sum_{m=-l}^l y^*_{lm}(x) y_{lm}(y) = Omega_pix^{-1} delta_{xy}
!
!=================================================================
! It should be noted that the latter condition is satisfied when
! the sum is taken from l=0. DO NOT EVER REMOVE l=0 and l=1.
! Removal of these two modes make a matrix y^*_i y_j SINGULAR!!
! (But this is true for temperature only. Polarization does not
!  have monopole and dipole by definition.)
!
! 12/17/03 EK (MODIFIED DNS)
!=================================================================
!
USE fitstools
USE alm_tools
use wmap_options, only : WMAP_data_dir

  IMPLICIT NONE
!
  INTEGER, INTENT(in)  :: nsmax,nlmax
  COMPLEX, INTENT(out) :: ylm(0:12*nsmax**2-1,(nlmax+1)*(nlmax+2)/2)

  REAL    :: map(0:12*nsmax**2-1,1:3)
  COMPLEX :: alm(1:3,0:nlmax,0:nlmax)
  INTEGER :: ipx,l,m,i,nterms,ih,it,io
  REAL :: omega_pix
  DOUBLE PRECISION, ALLOCATABLE :: w8ring(:,:)
  REAL, ALLOCATABLE :: dw8(:,:)
  CHARACTER*120 :: filename
!
  omega_pix = 3.1415926535/(3.*nsmax**2)
  nterms = (nlmax+1)*(nlmax+2)/2
  ALLOCATE(w8ring(1:2*nsmax,1))
  ALLOCATE(dw8(1:2*nsmax,1))
  ih = nsmax/100
  it = nsmax/10-10*ih
  io = nsmax -100*ih-10*it

  filename = trim(WMAP_data_dir)//'healpix_data/weight_ring_n00'//char(48+ih)//char(48+it)&
	& //char(48+io)//'.fits'
  call input_map(filename,dw8,2*nsmax,1)
  w8ring = 1.d0!+dw8
  do ipx = 0,12*nsmax**2-1

     map = 0.
     map(ipx,1) = 1.
     call map2alm(nsmax,nlmax,nlmax,map,alm,-1.d0,w8ring)
     i = 0
     do l = 0,nlmax
        do m = 0,l
           i = i+1
           ylm(ipx,i) = alm(1,l,m)/omega_pix
        end do
     enddo
   end do
!
   ylm = CONJG(ylm)
   DEALLOCATE(w8ring,dw8)
END SUBROUTINE generate_ylm

!
!
!
SUBROUTINE generate_plxy_matrix(nsmax,nlmax,Plxy)
! EK 12/13/03 modified DNS
!
! This subroutine calculates ``Legendre ! polynomials'':
!
! plxy(l,x,y) = Sum_m Y_lm(x) Y^*_lm(y)
!
! Note that
!
! Sum_m Y_lm(x) Y^*_lm(y) = Sum_{m>=0} Y_lm(x) Y^*_lm(y) 
!                          +Sum_{m<0} Y_{l-m}(x) Y^*_{l-m}(y)
!                         = Sum_{m>=0} Y_lm(x) Y^*_lm(y) 
!                          +Sum_{m<0} Y^*_lm(x) Y_lm(y)
! etc.
!
!
! version 2: now compute plxy(l) for l>=0 rather than l>=2, following
!          revision to `generate_ylm.f90'.
! 12/17/03 EK
!
use alm_tools
  IMPLICIT none
  INTEGER, intent(IN) :: nsmax,nlmax
  COMPLEX, ALLOCATABLE :: ylm(:,:)
  REAL, INTENT(OUT) :: Plxy(0:nlmax,0:12*nsmax*nsmax-1,0:12*nsmax*nsmax-1)

  INTEGER :: i,l,m,ip,jp,np
  np = 12*nsmax**2

  ALLOCATE(ylm(0:12*nsmax**2-1,(nlmax+1)*(nlmax+2)/2))

  call generate_ylm(nsmax,nlmax,ylm)

  do ip=0,np-1
     do jp=0,np-1
        i=0
        do l=0,nlmax
           i=i+1 ! m=0
           Plxy(l,ip,jp) = Ylm(ip,i)*conjg(Ylm(jp,i))
           do m=1,l
              i=i+1 ! m>0 plus m<0
              Plxy(l,ip,jp) = Plxy(l,ip,jp) &
                   + Ylm(ip,i)*conjg(Ylm(jp,i)) + Ylm(jp,i)*conjg(Ylm(ip,i))
           enddo
        enddo
     enddo
  enddo
 
  DEALLOCATE(ylm)
 
  RETURN
  END SUBROUTINE generate_plxy_matrix

  SUBROUTINE DERIV(t,t_deriv,nsmax)
use pix_tools
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: nsmax
  REAL, INTENT(IN) :: t(0:12*nsmax*nsmax-1)
  REAL, INTENT(OUT) :: t_deriv(0:12*nsmax*nsmax-1)
  INTEGER :: ipix,n(8),nneigh
  REAL :: T_AV

  DO ipix = 0,12*nsmax*nsmax-1
	CALL NEIGHBOURS_NEST(nsmax,ipix,n,nneigh)
	T_AV = SUM(t(n(1:nneigh)))/float(nneigh)
	t_deriv(ipix) = (t(ipix) - t_av)
  END DO
  t_deriv = 2*t_deriv*12*nsmax*nsmax/(4.*3.1415927)
  END SUBROUTINE DERIV

    SUBROUTINE COMPUTE_POWER(alm,blm,nlmax,cl)
    IMPLICIT NONE
    INTEGER :: nlmax !jxd
    COMPLEX, DIMENSION(1,0:nlmax,0:nlmax) :: alm,blm
    DOUBLE PRECISION :: cl(0:nlmax)
!    INTEGER :: nlmax,l,m  !jxd
    INTEGER :: l,m 
       DO l = 2,nlmax
 		cl(l) = alm(1,l,0)*blm(1,l,0)
		do m = 1,l
			cl(l) = cl(l)+alm(1,l,m)*CONJG(blm(1,l,m))+&
					CONJG(alm(1,l,m))*blm(1,l,m)
		end do
		cl(l) = cl(l)/(2*l+1)
       END DO
    RETURN
    END SUBROUTINE COMPUTE_POWER

    SUBROUTINE CONVERT_MAP_RING(t_ring,t_nest,nsmax)
    use pix_tools
    IMPLICIT NONE
    INTEGER :: nsmax
    REAL, DIMENSION(0:12*nsmax*nsmax-1) :: t_nest,t_ring
    INTEGER :: ipix_nest,ipix_ring
    DO ipix_nest = 0,12*nsmax*nsmax-1
	call nest2ring(nsmax,ipix_nest,ipix_ring)
	t_nest(ipix_nest) = t_ring(ipix_ring) 
    END DO
    RETURN
    END SUBROUTINE CONVERT_MAP_RING
    SUBROUTINE CONVERT_MAP(t_nest,t_ring,nsmax)
     use pix_tools
    IMPLICIT NONE
    INTEGER :: nsmax
    REAL, DIMENSION(0:12*nsmax*nsmax-1) :: t_nest,t_ring
    INTEGER :: ipix_nest,ipix_ring
    DO ipix_nest = 0,12*nsmax*nsmax-1
	call nest2ring(nsmax,ipix_nest,ipix_ring)
	t_ring(ipix_ring) = t_nest(ipix_nest)
    END DO
    RETURN
    END SUBROUTINE CONVERT_MAP
END MODULE tlike
